@@ -1,4 +1,5 @@
-"""Create a wheel (.whl) distribution.
+"""
+Create a wheel (.whl) distribution.
 
 A wheel is a built archive format.
 """
@@ -8,29 +9,32 @@ import hashlib
 import os
 import subprocess
 import warnings
-
-try:
-    import sysconfig
-except ImportError:  # pragma nocover
-    # Python < 2.7
-    import distutils.sysconfig as sysconfig
-
-import pkg_resources
-from pkg_resources import safe_name, safe_version
-
-from shutil import rmtree
+import shutil
+import json
+import sys
+import re
 from email.generator import Generator
-
-from distutils.util import get_platform
 from distutils.core import Command
 from distutils.sysconfig import get_python_version
-
 from distutils import log as logger
-import shutil
+from shutil import rmtree
 
-from .util import get_abbr_impl, get_impl_ver, native, open_for_csv
+import pkg_resources
+
+from .pep425tags import get_abbr_impl, get_impl_ver, get_abi_tag, get_platform
+from .util import native, open_for_csv
 from .archive import archive_wheelfile
 from .pkginfo import read_pkg_info, write_pkg_info
+from .metadata import pkginfo_to_dict
+from . import pep425tags, metadata
+from . import __version__ as wheel_version
+
+
+safe_name = pkg_resources.safe_name
+safe_version = pkg_resources.safe_version
+
+PY_LIMITED_API_PATTERN = r'cp3\d'
+
 
 def safer_name(name):
     return safe_name(name).replace('-', '_')
@@ -65,24 +69,43 @@ class bdist_wheel(Command):
                     ('group=', 'g',
                      "Group name used when creating a tar file"
                      " [default: current group]"),
+                    ('universal', None,
+                     "make a universal wheel"
+                     " (default: false)"),
+                    ('python-tag=', None,
+                     "Python implementation compatibility tag"
+                     " (default: py%s)" % get_impl_ver()[0]),
+                    ('build-number=', None,
+                     "Build number for this particular version. "
+                     "As specified in PEP-0427, this must start with a digit. "
+                     "[default: None]"),
+                    ('py-limited-api=', None,
+                     "Python tag (cp32|cp33|cpNN) for abi3 wheel tag"
+                     " (default: false)"),
                     ]
 
-    boolean_options = ['keep-temp', 'skip-build', 'relative']
+    boolean_options = ['keep-temp', 'skip-build', 'relative', 'universal']
 
     def initialize_options(self):
         self.bdist_dir = None
         self.data_dir = None
         self.plat_name = None
+        self.plat_tag = None
         self.format = 'zip'
         self.keep_temp = False
         self.dist_dir = None
         self.distinfo_dir = None
         self.egginfo_dir = None
-        self.root_is_purelib = None
+        self.root_is_pure = None
         self.skip_build = None
         self.relative = False
         self.owner = None
         self.group = None
+        self.universal = False
+        self.python_tag = 'py' + get_impl_ver()[0]
+        self.build_number = None
+        self.py_limited_api = False
+        self.plat_name_supplied = False
 
     def finalize_options(self):
         if self.bdist_dir is None:
@@ -90,42 +113,87 @@ class bdist_wheel(Command):
             self.bdist_dir = os.path.join(bdist_base, 'wheel')
 
         self.data_dir = self.wheel_dist_name + '.data'
+        self.plat_name_supplied = self.plat_name is not None
 
         need_options = ('dist_dir', 'plat_name', 'skip_build')
 
         self.set_undefined_options('bdist',
                                    *zip(need_options, need_options))
 
-        self.root_is_purelib = self.distribution.is_pure()
+        self.root_is_pure = not (self.distribution.has_ext_modules()
+                                 or self.distribution.has_c_libraries())
+
+        if self.py_limited_api and not re.match(PY_LIMITED_API_PATTERN, self.py_limited_api):
+            raise ValueError("py-limited-api must match '%s'" % PY_LIMITED_API_PATTERN)
+
+        # Support legacy [wheel] section for setting universal
+        wheel = self.distribution.get_option_dict('wheel')
+        if 'universal' in wheel:
+            # please don't define this in your global configs
+            val = wheel['universal'][1].strip()
+            if val.lower() in ('1', 'true', 'yes'):
+                self.universal = True
+
+        if self.build_number is not None and not self.build_number[:1].isdigit():
+            raise ValueError("Build tag (build-number) must start with a digit.")
 
     @property
     def wheel_dist_name(self):
         """Return distribution full name with - replaced with _"""
-        return '-'.join((safer_name(self.distribution.get_name()),
-                         safer_version(self.distribution.get_version())))
+        components = (safer_name(self.distribution.get_name()),
+                      safer_version(self.distribution.get_version()))
+        if self.build_number:
+            components += (self.build_number,)
+        return '-'.join(components)
+
+    def get_tag(self):
+        # bdist sets self.plat_name if unset, we should only use it for purepy
+        # wheels if the user supplied it.
+        if self.plat_name_supplied:
+            plat_name = self.plat_name
+        elif self.root_is_pure:
+            plat_name = 'any'
+        else:
+            plat_name = self.plat_name or get_platform()
+            if plat_name in ('linux-x86_64', 'linux_x86_64') and sys.maxsize == 2147483647:
+                plat_name = 'linux_i686'
+        plat_name = plat_name.replace('-', '_').replace('.', '_')
+
+        if self.root_is_pure:
+            if self.universal:
+                impl = 'py2.py3'
+            else:
+                impl = self.python_tag
+            tag = (impl, 'none', plat_name)
+        else:
+            impl_name = get_abbr_impl()
+            impl_ver = get_impl_ver()
+            impl = impl_name + impl_ver
+            # We don't work on CPython 3.1, 3.0.
+            if self.py_limited_api and (impl_name + impl_ver).startswith('cp3'):
+                impl = self.py_limited_api
+                abi_tag = 'abi3'
+            else:
+                abi_tag = str(get_abi_tag()).lower()
+            tag = (impl, abi_tag, plat_name)
+            supported_tags = pep425tags.get_supported(
+                supplied_platform=plat_name if self.plat_name_supplied else None)
+            # XXX switch to this alternate implementation for non-pure:
+            if not self.py_limited_api:
+                assert tag == supported_tags[0], "%s != %s" % (tag, supported_tags[0])
+            assert tag in supported_tags, "would build wheel with unsupported tag {}".format(tag)
+        return tag
 
     def get_archive_basename(self):
         """Return archive name without extension"""
-        purity = self.distribution.is_pure()
-        impl_ver = get_impl_ver()
-        abi_tag = 'none'
-        plat_name = 'any'
-        impl_name = 'py'
-        if not purity:
-            plat_name = self.plat_name.replace('-', '_').replace('.', '_')
-            impl_name = get_abbr_impl()
-            # PEP 3149 -- no SOABI in Py 2
-            # For PyPy?
-            # "pp%s%s" % (sys.pypy_version_info.major,
-            # sys.pypy_version_info.minor)
-            abi_tag = sysconfig.get_config_vars().get('SOABI', abi_tag)
-            abi_tag = abi_tag.rsplit('-', 1)[-1]
-        archive_basename = "%s-%s%s-%s-%s" % (
+
+        impl_tag, abi_tag, plat_tag = self.get_tag()
+
+        archive_basename = "%s-%s-%s-%s" % (
             self.wheel_dist_name,
-            impl_name,
-            impl_ver,
+            impl_tag,
             abi_tag,
-            plat_name)
+            plat_tag)
         return archive_basename
 
     def run(self):
@@ -142,6 +210,12 @@ class bdist_wheel(Command):
         install.skip_build = self.skip_build
         install.warn_dir = False
 
+        # A wheel without setuptools scripts is more cross-platform.
+        # Use the (undocumented) `no_ep` option to setuptools'
+        # install_scripts command to avoid creating entry point scripts.
+        install_scripts = self.reinitialize_command('install_scripts')
+        install_scripts.no_ep = True
+
         # Use a custom scheme for the archive, because we have to decide
         # at installation time which scheme to use.
         for key in ('headers', 'scripts', 'data', 'purelib', 'platlib'):
@@ -154,21 +228,15 @@ class bdist_wheel(Command):
         if os.name == 'nt':
             # win32 barfs if any of these are ''; could be '.'?
             # (distutils.command.install:change_roots bug)
-            basedir_observed = os.path.join(self.data_dir, '..')
+            basedir_observed = os.path.normpath(os.path.join(self.data_dir, '..'))
             self.install_libbase = self.install_lib = basedir_observed
 
-        if self.root_is_purelib:
-            setattr(install,
-                    'install_purelib',
-                    basedir_observed)
-        else:
-            setattr(install,
-                    'install_platlib',
-                    basedir_observed)
+        setattr(install,
+                'install_purelib' if self.root_is_pure else 'install_platlib',
+                basedir_observed)
 
         logger.info("installing to %s", self.bdist_dir)
-        if False:
-            self.fixup_data_files()
+
         self.run_command('install')
 
         archive_basename = self.get_archive_basename()
@@ -187,9 +255,6 @@ class bdist_wheel(Command):
                                          '%s.dist-info' % self.wheel_dist_name)
         self.egg2dist(self.egginfo_dir,
                       self.distinfo_dir)
-                
-        metadata_path = os.path.join(self.distinfo_dir, 'METADATA')
-        self.add_requirements(metadata_path)
 
         self.write_wheelfile(self.distinfo_dir)
 
@@ -199,7 +264,7 @@ class bdist_wheel(Command):
         if not os.path.exists(self.dist_dir):
             os.makedirs(self.dist_dir)
         wheel_name = archive_wheelfile(pseudoinstall_root, archive_root)
-        
+
         # Sign the archive
         if 'WHEEL_TOOL' in os.environ:
             subprocess.call([os.environ['WHEEL_TOOL'], 'sign', wheel_name])
@@ -214,30 +279,27 @@ class bdist_wheel(Command):
             else:
                 rmtree(self.bdist_dir)
 
-    def write_wheelfile(self, wheelfile_base, generator='bdist_wheel'):
+    def write_wheelfile(self, wheelfile_base, generator='bdist_wheel (' + wheel_version + ')'):
         from email.message import Message
         msg = Message()
-        msg['Wheel-Version'] = '0.1'  # of the spec
+        msg['Wheel-Version'] = '1.0'  # of the spec
         msg['Generator'] = generator
-        msg['Root-Is-Purelib'] = str(self.root_is_purelib).lower()
+        msg['Root-Is-Purelib'] = str(self.root_is_pure).lower()
+        if self.build_number is not None:
+            msg['Build'] = self.build_number
+
+        # Doesn't work for bdist_wininst
+        impl_tag, abi_tag, plat_tag = self.get_tag()
+        for impl in impl_tag.split('.'):
+            for abi in abi_tag.split('.'):
+                for plat in plat_tag.split('.'):
+                    msg['Tag'] = '-'.join((impl, abi, plat))
+
         wheelfile_path = os.path.join(wheelfile_base, 'WHEEL')
         logger.info('creating %s', wheelfile_path)
         with open(wheelfile_path, 'w') as f:
             Generator(f, maxheaderlen=0).flatten(msg)
 
-    def fixup_data_files(self):
-        """Put all resources in a .data directory"""
-        # (only useful in the packaging/distutils2 version of bdist_wheel)
-        data_files = {}
-        for k, v in self.distribution.data_files.items():
-            # {dist-info} is already in our directory tree
-            if v.startswith('{') and not v.startswith('{dist-info}'):
-                # XXX assert valid (in sysconfig.get_paths() or
-                # 'distribution.name')
-                data_files[k] = os.path.join(self.data_dir,
-                                             v.replace('{', '').replace('}', ''))
-        self.distribution.data_files.update(data_files)
-
     def _ensure_relative(self, path):
         # copied from dir_util, deleted
         drive, path = os.path.splitdrive(path)
@@ -245,49 +307,30 @@ class bdist_wheel(Command):
             path = drive + path[1:]
         return path
 
-    def _to_requires_dist(self, requirement):
-        requires_dist = []
-        for op, ver in requirement.specs:
-            # PEP 345 specifies but does not use == as part of a version spec
-            if op == '==':
-                op = ''
-            requires_dist.append(op + ver)
-        if not requires_dist:
-            return ''
-        return " (%s)" % ','.join(requires_dist)
-
     def _pkginfo_to_metadata(self, egg_info_path, pkginfo_path):
-        pkg_info = read_pkg_info(pkginfo_path)
-        pkg_info.replace_header('Metadata-Version', '1.3')
-        requires_path = os.path.join(egg_info_path, 'requires.txt')
-        if os.path.exists(requires_path):
-            requires = open(requires_path).read()
-            for extra, reqs in pkg_resources.split_sections(requires):
-                condition = ''
-                if extra:
-                    pkg_info['Provides-Extra'] = extra
-                    condition = '; extra == %s' % repr(extra)
-                for req in reqs:
-                    parsed_requirement = pkg_resources.Requirement.parse(req)
-                    spec = self._to_requires_dist(parsed_requirement)
-                    pkg_info['Requires-Dist'] = parsed_requirement.key + \
-                        spec + condition
-        return pkg_info
-    
+        return metadata.pkginfo_to_metadata(egg_info_path, pkginfo_path)
+
+    def license_file(self):
+        """Return license filename from a license-file key in setup.cfg, or None."""
+        metadata = self.distribution.get_option_dict('metadata')
+        if 'license_file' not in metadata:
+            return None
+        return metadata['license_file'][1]
+
     def setupcfg_requirements(self):
-        """Generate requirements from setup.cfg as 
+        """Generate requirements from setup.cfg as
         ('Requires-Dist', 'requirement; qualifier') tuples. From a metadata
         section in setup.cfg:
-        
+
         [metadata]
         provides-extra = extra1
             extra2
         requires-dist = requirement; qualifier
             another; qualifier2
             unqualified
-            
+
         Yields
-        
+
         ('Provides-Extra', 'extra1'),
         ('Provides-Extra', 'extra2'),
         ('Requires-Dist', 'requirement; qualifier'),
@@ -297,30 +340,32 @@ class bdist_wheel(Command):
         metadata = self.distribution.get_option_dict('metadata')
 
         # our .ini parser folds - to _ in key names:
-        for key, title in (('provides_extra', 'Provides-Extra'), 
+        for key, title in (('provides_extra', 'Provides-Extra'),
                            ('requires_dist', 'Requires-Dist')):
-            if not key in metadata:
+            if key not in metadata:
                 continue
             field = metadata[key]
             for line in field[1].splitlines():
                 line = line.strip()
                 if not line:
                     continue
-                yield (title, line) 
-    
+                yield (title, line)
+
     def add_requirements(self, metadata_path):
         """Add additional requirements from setup.cfg to file metadata_path"""
         additional = list(self.setupcfg_requirements())
-        if not additional: return        
+        if not additional:
+            return
+
         pkg_info = read_pkg_info(metadata_path)
         if 'Provides-Extra' in pkg_info or 'Requires-Dist' in pkg_info:
             warnings.warn('setup.cfg requirements overwrite values from setup.py')
             del pkg_info['Provides-Extra']
-            del pkg_info['Requires-Dist'] 
+            del pkg_info['Requires-Dist']
         for k, v in additional:
             pkg_info[k] = v
         write_pkg_info(metadata_path, pkg_info)
-        
+
     def egg2dist(self, egginfo_path, distinfo_path):
         """Convert an .egg-info directory into a .dist-info directory"""
         def adios(p):
@@ -359,47 +404,79 @@ class bdist_wheel(Command):
 
             # ignore common egg metadata that is useless to wheel
             shutil.copytree(egginfo_path, distinfo_path,
-                            ignore=lambda x, y: set(('PKG-INFO', 
-                                                     'requires.txt',
-                                                     'SOURCES.txt',
-                                                     'top_level.txt',
-                                                     'not-zip-safe',)))
-            
+                            ignore=lambda x, y: {'PKG-INFO', 'requires.txt', 'SOURCES.txt',
+                                                 'not-zip-safe'}
+                            )
+
             # delete dependency_links if it is only whitespace
-            dependency_links = os.path.join(distinfo_path, 'dependency_links.txt')
-            if not open(dependency_links, 'r').read().strip(): 
-                adios(dependency_links)        
+            dependency_links_path = os.path.join(distinfo_path, 'dependency_links.txt')
+            with open(dependency_links_path, 'r') as dependency_links_file:
+                dependency_links = dependency_links_file.read().strip()
+            if not dependency_links:
+                adios(dependency_links_path)
 
         write_pkg_info(os.path.join(distinfo_path, 'METADATA'), pkg_info)
 
+        # XXX deprecated. Still useful for current distribute/setuptools.
+        metadata_path = os.path.join(distinfo_path, 'METADATA')
+        self.add_requirements(metadata_path)
+
+        # XXX intentionally a different path than the PEP.
+        metadata_json_path = os.path.join(distinfo_path, 'metadata.json')
+        pymeta = pkginfo_to_dict(metadata_path,
+                                 distribution=self.distribution)
+
+        if 'description' in pymeta:
+            description_filename = 'DESCRIPTION.rst'
+            description_text = pymeta.pop('description')
+            description_path = os.path.join(distinfo_path,
+                                            description_filename)
+            with open(description_path, "wb") as description_file:
+                description_file.write(description_text.encode('utf-8'))
+            pymeta['extensions']['python.details']['document_names']['description'] = \
+                description_filename
+
+        # XXX heuristically copy any LICENSE/LICENSE.txt?
+        license = self.license_file()
+        if license:
+            license_filename = 'LICENSE.txt'
+            shutil.copy(license, os.path.join(self.distinfo_dir, license_filename))
+            pymeta['extensions']['python.details']['document_names']['license'] = license_filename
+
+        with open(metadata_json_path, "w") as metadata_json:
+            json.dump(pymeta, metadata_json, sort_keys=True)
+
         adios(egginfo_path)
 
     def write_record(self, bdist_dir, distinfo_dir):
-        from wheel.util import urlsafe_b64encode
+        from .util import urlsafe_b64encode
 
         record_path = os.path.join(distinfo_dir, 'RECORD')
         record_relpath = os.path.relpath(record_path, bdist_dir)
 
         def walk():
             for dir, dirs, files in os.walk(bdist_dir):
-                for f in files:
+                dirs.sort()
+                for f in sorted(files):
                     yield os.path.join(dir, f)
 
         def skip(path):
-            return (path.endswith('.pyc')
-                    or path.endswith('.pyo') or path == record_relpath)
+            """Wheel hashes every possible file."""
+            return (path == record_relpath)
 
-        writer = csv.writer(open_for_csv(record_path, 'w+'))
-        for path in walk():
-            relpath = os.path.relpath(path, bdist_dir)
-            if skip(relpath):
-                hash = ''
-                size = ''
-            else:
-                data = open(path, 'rb').read()
-                digest = hashlib.sha256(data).digest()
-                hash = 'sha256='+native(urlsafe_b64encode(digest))
-                size = len(data)
-            record_path = os.path.relpath(
-                path, bdist_dir).replace(os.path.sep, '/')
-            writer.writerow((record_path, hash, size))
+        with open_for_csv(record_path, 'w+') as record_file:
+            writer = csv.writer(record_file)
+            for path in walk():
+                relpath = os.path.relpath(path, bdist_dir)
+                if skip(relpath):
+                    hash = ''
+                    size = ''
+                else:
+                    with open(path, 'rb') as f:
+                        data = f.read()
+                    digest = hashlib.sha256(data).digest()
+                    hash = 'sha256=' + native(urlsafe_b64encode(digest))
+                    size = len(data)
+                record_path = os.path.relpath(
+                    path, bdist_dir).replace(os.path.sep, '/')
+                writer.writerow((record_path, hash, size))
